/* str.c -- operations on strings.

   Copyright (C) 1997 Free Software Foundation, Inc.
   Written by Thomas Morgan <tmorgan@gnu.ai.mit.edu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* Although this file is small enough to fit easily inside others, it
   seems to be useful for any program so I find it convenient for it
   to be seperate.  It was written because I became fed up with:

   * All string-handling code needing to worry about memory
   allocation.

   * Not being able to deal with NUL characters easily.

   * Sending three variables from function to function just for one
   string.  (They are now contained in a `str_t' type.)  */


/* Local headers.  */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "str.h"

/* System headers.  */

#include <errno.h>

#ifdef HAVE_STRING_H
#include <string.h>
#else /* not HAVE_STRING_H */
#include <strings.h>
#endif /* not HAVE_STRING_H */

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

/* These should be provided by the caller.  */
extern void *xmalloc PARAMS ((size_t));
extern void *xrealloc PARAMS ((void *, size_t));

static int safe_read PARAMS ((int, char *, int));

/* The name of the executable this process comes from.  This should be
   set by the caller.  */
extern char *program_name;

/* Whether we're reading from the terminal.  This should be set by the
   caller.  */
extern int interactive;


/* Initialize (or reinitialize) the string STR for use.  STR may be
   NULL; it will be changed.  Return the (sometimes new) pointer.  */

str_t *
str_make (str_t *str)
{
  if (str == NULL)
    {
      str = xmalloc (sizeof (str_t));
      str->str = xmalloc (str->mem = CHUNK);
    }
  else
     str->str = xrealloc (str->str, str->mem = CHUNK);
  str->len = 0;

  return str;
}

/* Free all memory associated with `str'.  */

void
str_free (str_t *str)
{
  if (str == NULL)
    return;
  free (str->str);
  free (str);
}

/* Append the character C to the string STR (create `*str' with
   `str_make').  */

void
str_add_char (str_t *str, char c)
{
  if (str == NULL || str->str == NULL)
    str = str_make (str);

  if (++(str->len) > str->mem)
    str->str = xrealloc (str->str, str->mem += CHUNK);

  str->str[str->len - 1] = c;
}

/* Append the string STR2 to the string STR1.  Both should be
   created with `str_make'.  */

void
str_add_str (str_t *str1, str_t *str2)
{
  int pos;

  if (str2 == NULL || str2->str == NULL)
    return;
  if (str1 == NULL || str1->str == NULL)
    str1 = str_make (str1);

  for (pos = 0; pos < str2->len; pos++)
    str_add_char (str1, str2->str[pos]);
}

/* Copy a newline-terminated line from STREAM to the string STR
   (create `str' with `str_make').  Return `ADD_LINE_OK' if
   successful, `ADD_LINE_EOF' if an EOF was gotten, or `ADD_LINE_ERR'
   in event of an error.  */

int
str_add_line (str_t *str, FILE *stream)
{
  if (str == NULL || str->str == NULL)
    str = str_make (str);
  if (stream == NULL)
    return ADD_LINE_ERR;

  while (1)
    {
      register int c = getc (stream);

      if (c == EOF || ferror (stream))
	return ADD_LINE_EOF;
      str_add_char (str, c);
      if (c == '\n')
	break;
    }

  return ADD_LINE_OK;
}

/* Copy a newline-terminated line from FILE_DESC to the string STR
   (create `str' with `str_make').  Return `ADD_LINE_OK' if
   successful, `ADD_LINE_EOF' if an EOF was gotten, or `ADD_LINE_ERR'
   in event of an error.  FIXME: should not read only one character at
   a time.  */

int
str_add_line_from_desc (str_t *str, int file_desc)
{
  int nchars = 0;
  char c = 0;

  if (str == NULL || str->str == NULL)
    str = str_make (str);

  while (1)
    {
      nchars = safe_read (file_desc, &c, 1);

      if (!nchars)
	return ADD_LINE_EOF;
      if (nchars < 0)
	return ADD_LINE_ERR;

      str_add_char (str, c);
      if (c == '\n')
	break;
    }

  return ADD_LINE_OK;
}

/* Convert the NUL-terminated character array *NSTR to a string
   structure and return it.  */

str_t *
nstr_to_str (char *nstr)
{
  str_t *str = str_make (0);
  int pos;

  if (nstr == NULL)
    return str;
  for (pos = 0; nstr[pos]; pos++)
    str_add_char (str, nstr[pos]);

  return str;
}

/* Convert the string STR (create `str' with `str_make') to a
   NUL-terminated character array and return it.  */

char *
str_to_nstr (str_t *str)
{
  char *nstr;
  int pos;

  if (str == NULL || str->str == NULL)
    return NULL;

  nstr = xmalloc (str->len + 1);

  for (pos = 0; pos < str->len; pos++)
    {
      if (str->str[pos] == '\0')
	/* NUL-terminated strings dislike having NULs in their
	   content, for obvious reasons.  I suppose the best thing is
	   to give them a space instead.  */
	str->str[pos] = ' ';

      nstr[pos] = str->str[pos];
    }

  nstr[pos] = 0;
  return nstr;
}

/* Convert the whole number NUM to a string structure and return it.  */

str_t *
int_to_str (int num)
{
  int lpos, rpos;		/* Positions in `str'.  */
  str_t *str = str_make (0);	/* Final string.  */

  /* Get the digits one at a time.  */
  do
    {
      str_add_char (str, num % 10 + '0');
      num /= 10;
    }
  while (num > 0);

  /* Reverse `str'.  */
  for (lpos = 0, rpos = str->len - 1; lpos < rpos; lpos++, rpos--)
    {
      char tmp = str->str[lpos];
      str->str[lpos] = str->str[rpos];
      str->str[rpos] = tmp;
    }

  return str;
}


/* Read LEN bytes at PTR from descriptor DESC, retrying if interrupted.
   Return the actual number of bytes read, zero for EOF, or negative
   for an error.  Written by the Free Software Foundation.  */

static int
safe_read (int desc, char *ptr, int len)
{
  int n_chars;

  if (len <= 0)
    return len;

#ifdef EINTR
  do
    {
      n_chars = read (desc, ptr, len);
    }
  while (n_chars < 0 && errno == EINTR);
#else
  n_chars = read (desc, ptr, len);
#endif

  return n_chars;
}
